﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {
  [Item("RSquaredEvaluator", "")]
  [StorableClass]
  public class RSquaredEvaluator : ParameterizedNamedItem, IGrammarEnumerationEvaluator {
    private readonly string OptimizeConstantsParameterName = "Optimize Constants";
    private readonly string ApplyLinearScalingParameterName = "Apply Linear Scaling";
    private readonly string ConstantOptimizationIterationsParameterName = "Constant Optimization Iterations";
    private readonly string RestartsParameterName = "Restarts";
    private readonly string SeedParameterName = "Seed"; // seed for the random number generator

    private readonly MersenneTwister random = new MersenneTwister();

    #region parameter properties
    public IFixedValueParameter<BoolValue> OptimizeConstantsParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[OptimizeConstantsParameterName]; }
    }

    public IFixedValueParameter<BoolValue> ApplyLinearScalingParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[ApplyLinearScalingParameterName]; }
    }

    public IFixedValueParameter<IntValue> ConstantOptimizationIterationsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[ConstantOptimizationIterationsParameterName]; }
    }

    private IFixedValueParameter<IntValue> RestartsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[RestartsParameterName]; }
    }

    private int Restarts {
      get { return RestartsParameter.Value.Value; }
      set { RestartsParameter.Value.Value = value; }
    }

    public bool OptimizeConstants {
      get { return OptimizeConstantsParameter.Value.Value; }
      set { OptimizeConstantsParameter.Value.Value = value; }
    }

    public bool ApplyLinearScaling {
      get { return ApplyLinearScalingParameter.Value.Value; }
      set { ApplyLinearScalingParameter.Value.Value = value; }
    }

    public int ConstantOptimizationIterations {
      get { return ConstantOptimizationIterationsParameter.Value.Value; }
      set { ConstantOptimizationIterationsParameter.Value.Value = value; }
    }
    #endregion

    private static readonly ISymbolicDataAnalysisExpressionTreeInterpreter expressionTreeLinearInterpreter = new SymbolicDataAnalysisExpressionTreeLinearInterpreter();

    public RSquaredEvaluator() {
      Parameters.Add(new FixedValueParameter<BoolValue>(OptimizeConstantsParameterName, "Run constant optimization in sentence evaluation.", new BoolValue(false)));
      Parameters.Add(new FixedValueParameter<BoolValue>(ApplyLinearScalingParameterName, "Apply linear scaling on the tree model during evaluation.", new BoolValue(false)));
      Parameters.Add(new FixedValueParameter<IntValue>(ConstantOptimizationIterationsParameterName, "Number of gradient descent iterations.", new IntValue(10)));
      Parameters.Add(new FixedValueParameter<IntValue>(RestartsParameterName, "Number of restarts for gradient descent.", new IntValue(10)));

      var seedParameter = new FixedValueParameter<IntValue>(SeedParameterName, "Seed value for random restarts.", new IntValue(0));
      seedParameter.Value.ValueChanged += (sender, args) => random.Seed((uint)seedParameter.Value.Value);
      random.Seed(0u);

      Parameters.Add(seedParameter);
    }

    [StorableConstructor]
    protected RSquaredEvaluator(bool deserializing) : base(deserializing) { }

    protected RSquaredEvaluator(RSquaredEvaluator original, Cloner cloner) : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new RSquaredEvaluator(this, cloner);
    }

    public double Evaluate(IRegressionProblemData problemData, Grammar grammar, SymbolList sentence) {
      var tree = grammar.ParseSymbolicExpressionTree(sentence);
      return Evaluate(problemData, tree);
    }

    public double Evaluate(IRegressionProblemData problemData, ISymbolicExpressionTree tree) {
      return Evaluate(problemData, tree, random, OptimizeConstants, ConstantOptimizationIterations, ApplyLinearScaling, Restarts);
    }

    public static double Evaluate(IRegressionProblemData problemData, ISymbolicExpressionTree tree, IRandom random, bool optimizeConstants = true, int maxIterations = 10, bool applyLinearScaling = false, int restarts = 1) {
      // we begin with an evaluation without constant optimization (relatively small speed penalty compared to const opt)
      // this value will be used as a baseline to decide if an improvement was achieved via const opt 
      double r2 = SymbolicRegressionSingleObjectivePearsonRSquaredEvaluator.Calculate(expressionTreeLinearInterpreter,
          tree,
          double.MinValue,
          double.MaxValue,
          problemData,
          problemData.TrainingIndices,
          applyLinearScaling: applyLinearScaling);

      // restart const opt and try to obtain an improved r2 value
      if (optimizeConstants) {
        int count = 0;
        double optimized = r2;
        do {
          foreach (var constantNode in tree.IterateNodesPrefix().OfType<ConstantTreeNode>()) {
            constantNode.ResetLocalParameters(random);
          }

          optimized = SymbolicRegressionConstantOptimizationEvaluator.OptimizeConstants(
            expressionTreeLinearInterpreter,
            tree,
            problemData,
            problemData.TrainingIndices,
            applyLinearScaling,
            maxIterations,
            false,
            double.MinValue,
            double.MaxValue,
            true);
        } while (optimized <= r2 && ++count < restarts);

        // do not update constants if quality is not improved
        if (optimized > r2) {
          r2 = optimized;

          // is this code really necessary ?
          foreach (var symbolicExpressionTreeNode in tree.IterateNodesPostfix()) {
            ConstantTreeNode constTreeNode = symbolicExpressionTreeNode as ConstantTreeNode;
            if (constTreeNode != null && constTreeNode.Value.IsAlmost(0.0)) {
              constTreeNode.Value = 0.0;
            }
          }
        }
      }
      return double.IsNaN(r2) || double.IsInfinity(r2) ? 0.0 : r2;
    }
  }
}
